"""HTTP client to the pixel API."""
from collections import defaultdict
from typing import Union, Optional

import aiohttp

from .canvas import Canvas
from .colours import Colour, parse_colour
from .ratelimits import RateLimiter


class Client:
    """HTTP client to the pixel API."""

    def __init__(
            self,
            token: str,
            base_url: str = 'https://pixels.pythondiscord.com/'):
        """Store the token and set up the client."""
        self.base_url = base_url
        self.headers = {'Authorization': 'Bearer ' + token}
        self.client = None
        # Cache canvas size, assuming it won't change.
        self.canvas_size = None
        self.ratelimits = defaultdict(RateLimiter)

    async def get_client(self) -> aiohttp.ClientSession:
        """Get or create the client session."""
        if (not self.client) or self.client.closed:
            self.client = aiohttp.ClientSession(headers=self.headers)
        return self.client

    async def request(
            self,
            method: str,
            endpoint: str,
            data: Optional[dict] = None,
            parse_json: bool = True) -> aiohttp.ClientResponse:
        """Make a call to an endpoint, respecting ratelimiting."""
        await self.ratelimits[endpoint].pause()
        client = await self.get_client()
        request = client.request(method, self.base_url + endpoint, json=data)
        async with request as response:
            self.ratelimits[endpoint].update(response.headers)
            if parse_json:
                return await response.json()
            else:
                return await response.read()

    async def put_pixel(
            self, x: int, y: int, colour: Union[int, str, Colour]) -> str:
        """Draw a pixel and return a message."""
        data = await self.request('POST', 'set_pixel', {
            'x': x,
            'y': y,
            'rgb': parse_colour(colour)
        })
        return data['message']

    async def get_canvas_size(self) -> tuple[int, int]:
        """Get the size of the canvas (with caching)."""
        if self.canvas_size:
            return self.canvas_size
        data = await self.request('GET', 'get_size')
        return data['width'], data['height']

    async def get_canvas(self) -> Canvas:
        """Request the entire canvas."""
        data = await self.request('GET', 'get_pixels', parse_json=False)
        size = await self.get_canvas_size()
        return Canvas(size, data)

    async def close(self):
        """Close the underlying session."""
        await self.client.close()



"""Utility for obeying ratelimits."""
import asyncio


class RateLimiter:
    """Ratelimiter for a specific endpoint."""

    def __init__(self):
        """Initialise fields to defaults."""
        self.ratelimited = True    # Not all endpoints are ratelimited.
        self.remaining = 1
        self.limit = None
        self.reset = None

    def update(self, headers: dict[str, int]):
        """Update the ratelimiter based on the latest headers."""
        if 'Requests-Remaining' not in headers:
            self.ratelimited = False
            return
        self.remaining = headers['Requests-Remaining']
        self.limit = headers['Requests-Limit']
        self.reset = headers['Requests-Reset']

    async def pause(self):
        """Pause before sending another request if necessary."""
        if self.remaining or not self.ratelimited:
            return
        await asyncio.sleep(self.reset)



"""Container for all the pixels on a canvas."""
import PIL.Image


class Canvas:
    """Container for all the pixels on a canvas."""

    def __init__(self, size: tuple[int, int], data: bytes):
        """Parse the raw canvas data."""
        self.width, self.height = size
        pixels = []
        for start_idx in range(0, len(data), 3):
            pixels.append(tuple(data[start_idx:start_idx + 3]))
        self.grid = [
            pixels[row * self.width:(row + 1) * self.width]
            for row in range(self.height)
        ]
        self.raw = data
        self.image = PIL.Image.frombytes('RGB', size, data)

    def show(self):
        """Display the image."""
        self.image.show()

    def save(self, path: str):
        """Save the image to a given file."""
        self.image.save(path)
